codeunit 62732 "MultiThreading Helper"
{
    /*
        AddThread(...) 
            Instruct to add a thread to be executed, could be called multiple time
        RunThreads(WaitForCompletion: Boolean; ShowGUI: Boolean; TimeOutMS: Integer) ErrorOccured: Boolean
            Launch execution of all requested threads
            WaitForCompletion : indicate to wait for all thread completion to continue the code
            ShowGUI : only if WaitforCompletion, display a window of thread status and progression
            TimeOutMS : milliseconds time out that will stop waiting and continue the code
            ErrorOccured : true if an error ocurred in any thread of timeout trigerred


        /!\ Limits : Business Central allow up to 100 child session, but only 5 are runned simulatenously on cloud, the others are queued
        https://learn.microsoft.com/en-us/dynamics365/business-central/dev-itpro/administration/operational-limits-online
    */

    var
        // Array usage ease compatibilty to older BC version
        SessionIDs: array[100] of Integer;
        ThreadCodeunit: array[100] of Integer;
        ThreadCount: Integer;
        ThreadCompany: array[100] of Text[50];
        ThreadNames: array[100] of Text[250];
        ThreadRecord: array[100] of Variant;

    procedure AddThread(CodeunitID: Integer) Sucess: Boolean
    var
        EmptyVariant: Variant;
    begin
        AddThread('Thread ' + format(ThreadCount + 1), CodeunitID, '', EmptyVariant);
    end;

    procedure AddThread(CodeunitID: Integer; CodeunitRecord: Variant) Sucess: Boolean
    begin
        AddThread('Thread ' + format(ThreadCount + 1), CodeunitID, '', CodeunitRecord);
    end;

    procedure AddThread(ThreadName: Text; CodeunitID: Integer) Sucess: Boolean
    var
        EmptyVariant: Variant;
    begin
        AddThread(ThreadName, CodeunitID, '', EmptyVariant);
    end;

    procedure AddThread(ThreadName: Text; CodeunitID: Integer; CodeunitRecord: Variant) Sucess: Boolean
    begin
        AddThread(ThreadName, CodeunitID, '', CodeunitRecord);
    end;

    procedure AddThread(ThreadName: Text; CodeunitID: Integer; DestinationCompanyName: Text[50]) Sucess: Boolean
    var
        EmptyVariant: Variant;
    begin
        AddThread(ThreadName, CodeunitID, DestinationCompanyName, EmptyVariant);
    end;


    procedure AddThread(ThreadName: Text; CodeunitID: Integer; DestinationCompanyName: Text[50]; CodeunitRecord: Variant) Sucess: Boolean
    var
        MaxConcurrentLbl: Label 'It is not possible to create more than 100 child sessions, please refer to : \ https://learn.microsoft.com/en-us/dynamics365/business-central/dev-itpro/administration/operational-limits-online';
    begin
        if ThreadCount >= 100 then error(MaxConcurrentLbl);
        ThreadCount += 1;
        ThreadCodeunit[ThreadCount] := CodeunitID;
        if DestinationCompanyName = '' then
            ThreadCompany[ThreadCount] := CompanyName
        else
            ThreadCompany[ThreadCount] := DestinationCompanyName;
        if ThreadName = '' then
            ThreadNames[ThreadCount] := format(ThreadCount)
        else
            ThreadNames[ThreadCount] := ThreadName;
        ThreadRecord[ThreadCount] := CodeunitRecord;
    end;

    procedure RunThreads(WaitForCompletion: Boolean; ShowGUI: Boolean; TimeOutMS: Integer) ErrorOccured: Boolean
    var
        i: Integer;
        EmptyVariant: Variant;
    begin
        for i := 1 to ThreadCount do begin
            if ThreadRecord[i].IsRecord() then
                StartSession(SessionIDs[i], ThreadCodeunit[i], ThreadCompany[i], ThreadRecord[i])
            else
                StartSession(SessionIDs[i], ThreadCodeunit[i], ThreadCompany[i]);
        end;
        if WaitForCompletion then
            exit(MonitorThreads(ShowGUI, SessionIDs, ThreadNames, TimeOutMS, 333));
    end;

    local procedure GetProgressBar(ProgressPercentage: Decimal) ProgressBar: Text
    var
        i: Integer;
    begin
        // Return a text progress bar using emote Eg : ⬜⬜⬜⬜⬜⬛⬛⬛⬛⬛⬛⬛⬛⬛⬛
        // ProgressPercentage must be betwen 0 and 1
        ProgressBar := '';
        for i := 1 to 20 do
            if (i < (ProgressPercentage * 20)) then
                ProgressBar += '⬜'
            else
                ProgressBar += '⬛';
    end;

    local procedure IsSessionEnded(SessionID: Integer): Boolean
    var
        SessionEvent: Record "Session Event";
    begin
        SessionEvent.SetFilter("Event Datetime", '>%1', CurrentDateTime - (1000 * 3600)); // less than 1h old
        SessionEvent.SetRange("Session ID", SessionID);
        SessionEvent.SetFilter("Event Type", '%1|%2|%3', SessionEvent."Event Type"::Logoff, SessionEvent."Event Type"::Stop, SessionEvent."Event Type"::Close);
        if not SessionEvent.IsEmpty then
            exit(true)
        else
            exit(false);
    end;

    local procedure IsSessionError(SessionID: Integer): Text[132]
    var
        SessionEvent: Record "Session Event";
    begin
        SessionEvent.SetFilter("Event Datetime", '>%1', CurrentDateTime - (1000 * 3600)); // less than 1h old
        SessionEvent.SetRange("Session ID", SessionID);
        SessionEvent.SetRange("Event Type", SessionEvent."Event Type"::Logoff);
        if SessionEvent.FindFirst() then
            exit(SessionEvent.Comment);
    end;

    local procedure IsSessionStarted(SessionID: Integer): Boolean
    var
        SessionEvent: Record "Session Event";
    begin
        SessionEvent.SetFilter("Event Datetime", '>%1', CurrentDateTime - (1000 * 3600)); // less than 1h old
        SessionEvent.SetRange("Session ID", SessionID);
        SessionEvent.SetFilter("Event Type", '%1|%2', SessionEvent."Event Type"::Logon, SessionEvent."Event Type"::Start);
        if not SessionEvent.IsEmpty then
            exit(true)
        else
            exit(false);
    end;

    local procedure MonitorThreads(ShowGUI: Boolean; var SessionsIDs: array[100] of Integer; SessionsNames: array[100] of Text[250]; TimeOutMS: Integer; RefreshRateMS: Integer) ErrorHappened: Boolean
    var
        AllFinished: Boolean;
        SessionIsFinished: array[100] of Boolean;
        SessionIsStarted: array[100] of Boolean;
        StartedDT: DateTime;
        Window: Dialog;
        i: Integer;
        SessionCount: Integer;
        SessionFinishedCount: Integer;
        FinishedLbl: Label 'The process has finished in: %1 \ %2';
        ProgressLbl: Label 'Multithread processing... \ Progression : #1### \ #2### \ #3###';
        StrThreadsStatusList: Text;
        CRLF: Text[2];
        // GUI purpose
        LoadingChar: Text[2];
        SessionError: array[100] of Text[132];
    begin
        if TimeOutMS = 0 then TimeOutMs := 1000 * 60 * 60 * 12; // Default Busienss Central childSession Timeout is 12h
        StartedDT := CurrentDateTime;

        // Init. GUI variables
        LoadingChar := '⚫';
        CRLF[1] := 13;
        CRLF[2] := 10;
        if GuiAllowed and ShowGUI then
            Window.Open(ProgressLbl);

        // Lookup array length
        for i := 1 to 100 do
            if SessionsIDs[i] <> 0 then
                SessionCount += 1
            else
                break;

        // While all threads are not finished
        AllFinished := false;
        while (not AllFinished) do begin
            // check time out
            if (currentdatetime - StartedDT) > TimeOutMS then begin
                ErrorHappened := true;
                break;
            end;

            if UpdateSessionsStatus(AllFinished, SessionFinishedCount, SessionsIDs, SessionIsStarted, SessionIsFinished, SessionError) then
                ErrorHappened := true;

            // Update Progression GUI
            if GuiAllowed and ShowGUI then begin
                Window.Update(1, format(SessionFinishedCount) + ' / ' + format(SessionCount));
                Window.Update(2, GetProgressBar(SessionFinishedCount / SessionCount));

                // Update Thread status list
                // Blinking for running thread
                if LoadingChar = '⚫' then
                    LoadingChar := '🟢'
                else
                    LoadingChar := '⚫';

                StrThreadsStatusList := '';
                // Loop threads
                for i := 1 to 100 do begin
                    if SessionsIDs[i] = 0 then break;
                    // Thread status builder
                    if not SessionIsStarted[i] then
                        StrThreadsStatusList += ' [⚫]'
                    else
                        if not SessionIsFinished[i] then
                            StrThreadsStatusList += ' [' + LoadingChar + ']'
                        else
                            if SessionError[i] <> '' then
                                StrThreadsStatusList += ' [❌]'
                            else
                                StrThreadsStatusList += ' [✔️]';
                    // Add thread name and eventual error message
                    StrThreadsStatusList += ' ' + SessionsNames[i] + ' - ' + SessionError[i] + CRLF;
                end;
                Window.Update(3, StrThreadsStatusList);
            end;
            sleep(RefreshRateMS);
        end;

        if GuiAllowed and ShowGUI then begin
            Window.Close();
            Message(FinishedLbl, CurrentDateTime - StartedDT, StrThreadsStatusList);
        end;
    end;

    local procedure UpdateSessionsStatus(var AllFinished: Boolean; var NumberOfSessionEnded: Integer; var SessionsIDs: array[100] of Integer; var SessionIsStarted: array[100] of Boolean; var SessionIsEnded: array[100] of Boolean; var SessionError: array[100] of Text[132]) ErrorOccured: Boolean
    var
        i: Integer;
    begin
        AllFinished := true;
        ErrorOccured := false;

        SelectLatestVersion();
        // Check each subsession status using session events (started, finished, error)
        for i := 1 to 100 do begin
            if SessionsIDs[i] = 0 then break;
            if not SessionIsEnded[i] then
                if not SessionIsStarted[i] then begin
                    // Check if started
                    if IsSessionStarted(SessionsIDs[i]) then
                        SessionIsStarted[i] := true;
                    AllFinished := false;
                end else
                    // Check if ended
                    if IsSessionEnded(SessionsIDs[i]) then begin
                        SessionIsEnded[i] := true;
                        NumberOfSessionEnded += 1;
                        // Check for error in session event log
                        SessionError[i] := IsSessionError(SessionsIDs[i]);
                        if SessionError[i] <> '' then ErrorOccured := true;
                    end else
                        AllFinished := false
        end;
    end;
}